import binascii
import os
import re
import subprocess
import sys
import datetime

try :
	#import colorama
	from colorama import Fore, Back, Style, init

except :
	print("Colorama is not installed! Use \"pip install colorama\" from your Python\Scripts folder.\n")
	sys.exit()

#colorama.init()
init()

def id_from_bin (id_data, output) :
	
	hex_bin  = binascii.hexlify(id_data).decode('utf-8').upper()
	id_hex   = binascii.hexlify(id_data[::-1]).decode('utf-8').upper()
	#id_hex   = id_hex[4:] + id_hex[:4]
	pci_ven  = id_hex[4:]
	pci_dev  = id_hex[:4]
	ven_dev  = pci_ven + "-" + pci_dev
	
	if output == "hexbin" :
		return hex_bin
	elif output == "hexrev" :
		return id_hex
	elif output == "id_list" :
		return pci_ven, pci_dev
	elif output == "all_list" :
		return ven_dev, pci_ven, pci_dev
	elif output == "hex" :
		return pci_ven + pci_dev
	else :
		return ven_dev

def rom_info(reading, get_info) :
	t_position = 0
	ifr_header = False
	rom_found  = False
	t_old_type = False
	t_rom_pat  = re.compile(br'\x55\xAA')
	mini_me    = True if get_info == "mini" else False
	basic_me   = True if get_info == "basic" else False
	
	## Checking for Nvidia IFR header
	nv_ifr_test = reading[0:4]
	
	if nv_ifr_test == b'NVGI' :
		ifr_header = True
		ifr_size   = int.from_bytes(reading[0x14:0x16], 'little')
		t_position = ifr_size
		
		if mini_me :
			print(Fore.GREEN + "Found Nvidia IFR header before ROM start, size 0x%X\n" % ifr_size + Fore.RESET)
			ifr_id = id_from_bin(reading[0xC:0x10], "string")
			print(Fore.GREEN + "ID of IFR header  = %s\n" % ifr_id + Fore.RESET)
	
	## Not really needed, but it is safe. The ROM should start at 00 or at ifr_size.
	while not rom_found :
		
		t_rom_match = t_rom_pat.search(reading, t_position)
		
		if t_rom_match is None :
			
			if basic_me :
				return False
			
			print(Fore.RED + "No ROM found!\n" + Fore.RESET)
			file_dec = "%s_decompr.bin" % file_rom
			
			if not os.path.isfile(file_dec) :
				print(Fore.RED + "Trying direct decompression...\n" + Fore.RESET)
				decomp = subprocess.call(["UEFIRomExtract", file_dir, file_dec], shell=True)
				print("")
				
				if os.path.isfile(file_dec) :
					print(Fore.YELLOW + "File " + file_dec + " was written! \n" + Fore.RESET)
			
			sys.exit()
		
		(t_start_match, t_end_match) = t_rom_match.span()
		t_start_pcir = int.from_bytes(reading[t_start_match + 0x18:t_start_match + 0x1A], 'little')
		#print("0x%0.2X - 0x%0.2X" % (t_start_match, t_start_pcir))
		
		if t_start_pcir == 0 :
			t_start_pnp = int.from_bytes(reading[t_start_match + 0x1A:t_start_match + 0x1C], 'little')
			t_pnp_off   = t_start_match + t_start_pnp
			t_pnp_str   = reading[t_pnp_off:t_pnp_off + 4]
			
			if t_pnp_str != b'$PnP' :
				#print("No PnP\n")
				t_position = t_start_match + 2
				continue
			
			t_start_pcir = 0x20
		
		t_pcir_off = t_start_match + t_start_pcir
		t_pcir_str = reading[t_pcir_off:t_pcir_off + 4]
		#print (t_pcir_str)
		
		if t_pcir_str != b'PCIR' :
			#print("No PCIR\n")
			t_position = t_start_match + 2
			continue
		
		## Only first image needed
		rom_found = True
	
	if basic_me :
		return True
	
	## Get some ROM related info
	t_pcir_id_bin = reading[t_pcir_off + 4:t_pcir_off + 8]
	#t_pcir_id_hex = binascii.hexlify(t_pcir_id_bin[::-1]).decode('utf-8').upper()
	t_ven_id, t_did_id = id_from_bin(t_pcir_id_bin, "id_list")
	
	if mini_me :
		## Only Nvidia IFR found so far as header before ROM
		if t_start_match and not ifr_header :
			print(Fore.MAGENTA + "Found Unknown header before ROM start, size 0x%X\n" % t_start_match + Fore.RESET)
		
		print(Style.BRIGHT + Fore.CYAN + "ID of ROM file    = %s-%s\n" % (t_ven_id, t_did_id)  + Fore.RESET + Style.NORMAL)
		
		if t_ven_id == "10DE" and t_did_id in ['0FF2', '11BF'] :
			print(Style.BRIGHT + Fore.YELLOW + "Nvidia GRID K1/K2 was detected! Using Multi-Display GOP.\n" + Fore.RESET + Style.NORMAL)
	
	## Get boundaries and sizes
	if reading[t_start_match + 4:t_start_match + 8] == b'\xF1\x0E\x00\x00' :
		## This part is only for extraction of single EFI ROM files
		t_rom_size  = 0
		t_rom_end   = t_start_match
		t_efi_found = True
		t_efi_begin = t_start_match
		
	else :
		## Calculate first ROM size. Sometimes the size from PCI DS is the right one.
		t_rom_size    = ord(reading[t_start_match + 2:t_start_match + 3]) * 0x200
		t_rom_size_ds = int.from_bytes(reading[t_pcir_off + 0x10:t_pcir_off + 0x12], 'little') * 0x200
		
		if t_rom_size != t_rom_size_ds :
			print(Fore.MAGENTA + "Different sizes in ROM header and PCI structure of Legacy ROM!\n" + Fore.RESET)
			t_rom_end_ds = t_start_match + t_rom_size_ds
			rom_test_ds  = reading[t_rom_end_ds:t_rom_end_ds + 2]
			
			if rom_test_ds in [b'\x55\xAA', b'\x56\x4E'] :
				t_rom_size = t_rom_size_ds
		
		## If all the images are contained in first ROM.
		#if reading[t_pcir_off + 0x20:t_pcir_off + 0x24] == b'NPDE' :
		#	t_npde_size = int.from_bytes(reading[t_pcir_off + 0x28:t_pcir_off + 0x2A], 'little') * 0x200
		#	
		#	if t_rom_size != t_npde_size :
		#		print(Fore.MAGENTA + "Different sizes in PCI structure and NPDE structure of Legacy ROM!\n" + Fore.RESET)
		#		## If the EFI image should be placed between Legacy and Special images, thus breaking the big container.
		#		t_rom_end_npde = t_start_match + t_npde_size
		#		rom_test_npde  = binascii.hexlify(reading[t_rom_end_npde:t_rom_end_npde + 2]).decode('utf-8').upper()
		#		if rom_test_npde in ['55AA', '564E'] :
		#			t_rom_size = t_npde_size
		
		t_rom_end   = t_start_match + t_rom_size
	
	## Not really needed, but it is safe. EFI should start at t_rom_end.
	## Assume that EFI follows OROM and that there is nothing in between.
	## The only structure found so far was Legacy ROM + EFI ROM, no other images in between.
	
	t_pat_efi   = re.compile(br'\x55\xAA..\xF1\x0E\x00\x00', re.DOTALL)
	t_match_efi = t_pat_efi.search(reading, t_rom_end)
	
	if t_match_efi is not None :
		(t_start_efi_match, t_end_efi_match) = t_match_efi.span()
		t_efi_found = True
		t_efi_begin = t_start_efi_match
		
		# Scan for special images between ROM and GOP
		while reading[t_rom_end:t_rom_end + 2] == b'\x56\x4E' :
			npds_ptr  = int.from_bytes(reading[t_rom_end + 0x18:t_rom_end + 0x1A], 'little')
			npds_off  = t_rom_end + npds_ptr
			npds_size = int.from_bytes(reading[npds_off + 0x10:npds_off + 0x12], 'little') * 0x200
			npds_type = "%02Xh" % ord(reading[npds_off + 0x14:npds_off + 0x15])
			t_rom_end += npds_size
			
			if mini_me :
				print(Style.BRIGHT + Fore.YELLOW + "Found special image %s between ROM and EFI!\n" % npds_type + Fore.RESET)
		
		# Skip intermediate images for other vendors. Added only for extraction. They will be reported later.
		while t_ven_id not in ['1002', '10DE'] and reading[t_rom_end:t_rom_end + 2] == b'\x55\xAA' :
			temp_pcir = int.from_bytes(reading[t_rom_end + 0x18:t_rom_end + 0x1A], 'little')
			temp_off  = t_rom_end + temp_pcir
			
			if reading[temp_off:temp_off + 4] != b'PCIR' or reading[t_rom_end + 4:t_rom_end + 8] == b'\xF1\x0E\x00\x00' :
				break
			
			temp_size    = ord(reading[t_rom_end + 2:t_rom_end + 3]) * 0x200
			tmp_size_ds  = int.from_bytes(reading[temp_off + 0x10:temp_off + 0x12], 'little') * 0x200
			
			if temp_size != tmp_size_ds :
				t_rom_end_ds = t_rom_end + tmp_size_ds
				rom_test_ds  = reading[t_rom_end_ds:t_rom_end_ds + 2]
				
				if rom_test_ds in [b'\x55\xAA', b'\x56\x4E'] :
					temp_size = tmp_size_ds
			
			t_rom_end += temp_size
				
		if t_efi_begin != t_rom_end :
			print(Style.BRIGHT + Fore.RED + "Data between ROM and EFI! Please report it\n" + Fore.RESET)
			sys.exit()
		
	else :
		#t_start_efi_match = 0
		t_efi_found = False
		t_efi_begin = t_rom_end
	
	if t_efi_found :
		t_efi_size = int.from_bytes(reading[t_efi_begin + 2:t_efi_begin + 4], 'little') * 0x200
		
		if mini_me :
			return (t_efi_found, t_efi_begin, t_efi_size)
	else :
		if mini_me :
			return (t_efi_found, 0, 0)
		
		## Check for a dummy EFI image. If it starts with 564E, it is a special Nvidia image.
		## If it is dummy EFI, the first two bytes are 77BB.
		
		if t_ven_id == "10DE" :
			
			test_image = reading[t_efi_begin:t_efi_begin + 2]
			
			## Test for dummy EFI in Nvidia.
			if len(test_image) == 2 and test_image != b'\x56\x4E' : ## 55AA or 77BB = old structure or dummy
				pci_ds_efi   = int.from_bytes(reading[t_efi_begin + 0x18:t_efi_begin + 0x1A], 'little')
				pcir_efi_off = t_efi_begin + pci_ds_efi
				t_struct     = reading[pcir_efi_off:pcir_efi_off + 4]
				
				## Old images have the special Nvidia images as normal ROMs
				if test_image == b'\x55\xAA' and t_struct == b'PCIR' :
					print(Style.BRIGHT + Fore.YELLOW + "Old structure was found in special images!\n" + Fore.RESET + Style.NORMAL)
					t_old_type = True
					t_efi_size = 0
				## A dummy EFI should have only one of the following two structures.
				elif test_image == b'\x77\xBB' and t_struct in [b'NPDS', b'RGIS'] :
					print(Style.BRIGHT + Fore.YELLOW + "Dummy EFI image was found!\n" + Fore.RESET + Style.NORMAL)
					t_efi_size = int.from_bytes(reading[pcir_efi_off + 0x10:pcir_efi_off + 0x12], 'little') * 0x200
				## Not old image, not dummy, not special image. What can it be?
				else :
					print(Style.BRIGHT + Fore.RED + "Strange Nvidia image was found after ROM! Please report it!\n" + 
					Fore.RESET + Style.NORMAL)
					t_efi_size = 0
			
			## No more images or Nvidia special image.
			else :
				t_efi_size = 0
		
		## Is AMD and AMD has no dummy EFI and no other images, from what I have seen.
		else :
			t_efi_size = 0
			
			mc_off_list = [0x1A000, 0x1B800, 0x1C000]
			
			for mc_off_idx in mc_off_list :
				
				mc_off     = t_start_match + mc_off_idx
			
				if reading[mc_off:mc_off + 4] == b'MCuC' :
					print(Style.BRIGHT + Fore.YELLOW + "AMD microcode was found at offset 0x%0.2X!\n" % mc_off + 
					Fore.RESET + Style.NORMAL)
					break
				
	return (t_old_type, t_start_match, t_pcir_off, t_pcir_id_bin, t_rom_size, t_efi_found, t_efi_begin, t_efi_size)

def efi_version(t_efi_dump) :
	t_gop_type = ""
	t_nv_type  = ""
	t_version  = ""
	t_efi_info_string = ""
	
	## AMD GOP
	pat_amd   = re.compile(br'\x44\x00\x72\x00\x69\x00\x76\x00\x65\x00\x72\x00\x20\x00\x52\x00\x65\x00\x76\x00') ## D.r.i.v.e.r. .R.e.v.
	match_amd = pat_amd.search(t_efi_dump)
	
	if match_amd is not None :
		(ver_start_match, ver_end_match) = match_amd.span()
		t_gop_type = "AMD"
		t_version  = lib_build = date = build = changelist = "0"
		bios_idtf_gop = bios_idtf_rom = 0
		
		t_version = t_efi_dump[ver_end_match + 2: ver_end_match + 0x1A].decode('utf-16', 'ignore').upper()
		
		if t_version[8:9] == "." :
			
			if t_version[:1] == "0" :
				t_version = t_version[:8]
			else :
				t_version = t_version[:10]
		
		verslen = 2 * len(t_version)
		lib     = ver_end_match + 4 + verslen
		test    = t_efi_dump[lib:lib + 6].decode('utf-16', 'ignore')
		
		if test == "Lib" : ## "4C0069006200"
			#print("Lib")
			lib_build = t_efi_dump[lib + 0x16:lib + 0x1E].decode('utf-16', 'ignore')
			date  = t_efi_dump[lib + 0x20:lib + 0x48].decode('utf-16', 'ignore')
		
		elif test.isalpha() :
			#print("None")
			lib_build = "----"
			date  = t_efi_dump[lib:lib + 0x28].decode('utf-16', 'ignore')
		
		elif test.isnumeric() :
			#print("Old")
			lib_build = t_efi_dump[lib:lib + 0xC].decode('utf-16', 'ignore')
			
			try :
				dot_off  = lib_build.index(".")
				date_off = lib + dot_off * 2 + 2
				lib_build = lib_build[:dot_off]
			except :
				date_off = lib + 0xE
				
			date  = t_efi_dump[date_off:date_off + 0x28].decode('utf-16', 'ignore')
		
		print(Style.BRIGHT + Fore.RED + "AMD GOP" + Fore.WHITE + " %s  " % t_version + 
		Fore.RED + "LibBuild" + Fore.WHITE + " %s  " % lib_build + 
		Fore.RED + "Dated:" + Fore.WHITE + " %s" % date.replace(".", " ") + 
		Fore.RESET + Style.NORMAL)
		
		t_efi_info_string = t_version + " - " + lib_build.replace("-", "*") + " - " + date.replace(".", " ")
		
		## BIOS_IDTF + AMD_Build + AMD_CL. Each one takes 0x18 bytes.
		pat_idtf   = re.compile(br'\x42\x49\x4F\x53\x5F\x49\x44\x54\x46') ## BIOS_IDTF
		match_idtf = pat_idtf.search(t_efi_dump)
		
		if match_idtf is not None :
			(idtf_start_match, idtf_end_match) = match_idtf.span()
			bios_idtf_gop = int.from_bytes(t_efi_dump[idtf_start_match + 0xA:idtf_start_match + 0xE], 'little')
		else :
			idtf_start_match = 0
		
		if idtf_start_match and t_efi_dump[idtf_start_match + 0x18:idtf_start_match + 0x21] == b'AMD_Build' :
			bd_start_match = idtf_start_match + 0x18
			build = int.from_bytes(t_efi_dump[bd_start_match + 0xA:bd_start_match + 0x18], 'little')
		else :
			pat_bd   = re.compile(br'\x41\x4D\x44\x5F\x42\x75\x69\x6C\x64') ## AMD_Build
			match_bd = pat_bd.search(t_efi_dump)
			
			if match_bd is not None :
				(bd_start_match, bd_end_match) = match_bd.span()
				build = int.from_bytes(t_efi_dump[bd_start_match + 0xA:bd_start_match + 0x18], 'little')
			else :
				bd_start_match = 0
		
		if bd_start_match and t_efi_dump[bd_start_match + 0x18:bd_start_match + 0x1E] == b'AMD_CL' :
			cl_start_match = bd_start_match + 0x18
			changelist = int.from_bytes(t_efi_dump[cl_start_match + 0xA:cl_start_match + 0x18], 'little')
		else :
			pat_cl   = re.compile(br'\x41\x4D\x44\x5F\x43\x4C') ## AMD_CL
			match_cl = pat_cl.search(t_efi_dump)
			
			if match_cl is not None :
				(cl_start_match, cl_end_match) = match_cl.span()
				changelist = int.from_bytes(t_efi_dump[cl_start_match + 0xA:cl_start_match + 0x18], 'little')
		
		# TODO Use ATOMBIOS Tables to get CRC offset.
		pat_lg   = re.compile(br'\x41(\x4D\x44|\x54\x49)\x20\x41\x54\x4F\x4D\x42\x49\x4F\x53\x00') ## AMD ATOMBIOS or ATI ATOMBIOS.
		match_lg = pat_lg.search(t_efi_dump)
		
		if match_lg is not None :
			(lg_start_match, lg_end_match) = match_lg.span()
			#print("\nAMD GOP has tables! Customized GOP with legacy parts!")
			bios_idtf_rom = int.from_bytes(t_efi_dump[lg_start_match + 0xD:lg_start_match + 0x11], 'little')
			#bios_idtf_rom = " ## Legacy BIOS_IDTF 0x" + bios_idtf_rom
		
			print(Style.BRIGHT + Fore.RED + "\nAMD_Build" + Fore.WHITE + " %s  " % build + 
			Fore.RED + "AMD_ChangeList" + Fore.WHITE + " %s  " % changelist + 
			Fore.RED + "GOP BIOS_IDTF" + Fore.WHITE + " 0x%0.8X " % bios_idtf_gop + 
			"\n\nAMD GOP has tables! Customized GOP with legacy parts!!" + 
			Fore.RED + "\n\nLegacy BIOS_IDTF" + Fore.WHITE + " 0x%0.8X " % bios_idtf_rom + 
			Fore.RESET + Style.NORMAL)
			
			t_efi_info_string += " - " + "%s" % build + " - " + "%s" % changelist + " - " + "0x%0.8X" % bios_idtf_gop + \
			" - " + "0x%0.8X" % bios_idtf_rom
		
		else :
			print(Style.BRIGHT + Fore.RED + "\nAMD_Build" + Fore.WHITE + " %s  " % build + 
			Fore.RED + "AMD_ChangeList" + Fore.WHITE + " %s  " % changelist + 
			Fore.RED + "GOP BIOS_IDTF" + Fore.WHITE + " 0x%0.8X " % bios_idtf_gop + 
			Fore.RESET + Style.NORMAL)
			
			t_efi_info_string += " - " + "%s" % build + " - " + "%s" % changelist + " - " + "0x%0.8X" % bios_idtf_gop
		
		## IDs in GOP. Works only with newer GOPs.
		
		# Alternative for ID search
		#pat_id_ptr   = re.compile(br'\x01\x01\x00{6}.{4}\x00{4}\x01\x02\x00{6}.{4}\x00{4}\x01\x03\x00{6}.{4}\x00{4}\x01\x04\x00{6}.{4}\x00{4}', re.DOTALL)
		#match_id_ptr = pat_id_ptr.search(t_efi_dump)
		#
		#if match_id_ptr is not None :
		#	(id_ptr_start_match, id_ptr_end_match) = match_id_ptr.span()
		#	id_start_match = int.from_bytes(t_efi_dump[id_ptr_start_match + 8:id_ptr_start_match + 0xC], 'little')
		
		pat_id   = re.compile(br'\x00{8}\x88\x68\x00{6}')
		match_id = pat_id.search(t_efi_dump)
		
		if match_id is not None :
			(id_start_match, id_end_match) = match_id.span()
			ids_list = ""
			step = id_start_match + 8
			
			while True :
				check_null = t_efi_dump[step + 2:step + 8]
				
				if check_null != b'\x00' * 6 : # "000000000000"
					break
				
				id = "1002-" + binascii.hexlify(t_efi_dump[step:step + 2][::-1]).decode('utf-8').upper()
				
				if id == "1002-0000" :
					step += 0x10
					continue
				
				if id == "1002-0101" :
					break
				
				name_off  = int.from_bytes(t_efi_dump[step + 8:step + 0xC], 'little')
				name_str  = ""
				name_step = name_off
				name_chr  = t_efi_dump[name_step:name_step + 1]
				
				while name_chr != b'\x00' :
					name_str  += name_chr.decode('utf-8', 'ignore')
					name_step += 1
					name_chr  = t_efi_dump[name_step:name_step + 1]
				
				ids_list += id + "  =  " + name_str + "\n"
				step     += 0x10
			
			with open("%s_temp\\AMD_GOP_%s_IDs.txt" % (file_rom, t_version), "a") as myfile :
				myfile.write(ids_list)
		
		## Names in GOP. Only AMD with GOPs newer than 1.34.0.0.0.
		#pat_name   = re.compile(br'\x41\x00\x6D\x00\x64\x00\x41\x00\x63\x00\x70\x00\x69\x00\x56\x00\x61\x00\x72\x00') ## A.m.d.A.c.p.i.V.a.r.
		#match_name = pat_name.search(t_efi_dump)
		#
		#if match_name is not None and int(t_version[2:4]) > 34 :
		#	#print("\nFound names")
		#	(name_start_match, name_end_match) = match_name.span()
		#	bin_test = t_efi_dump[name_end_match + 4:name_end_match + 0xE]
		#	str_test = bin_test.decode('utf-8', 'ignore')
		#	#print(bin_test)
		#	#print(str_test)
		#	
		#	if str_test == "CR has Bad" or bin_test == b'\x4C\x00\x4C\x00\x4C\x00\x4C\x00\x4C\x00' :
		#		#print("\nCR or LL")
		#		pat_name   = re.compile(br'\x00{4}\x65\x6E\x00{2}') ## ....en..
		#		match_name = pat_name.search(t_efi_dump)
		#		
		#		if match_name is not None :
		#			(name_start_match, name_end_match) = match_name.span()
		#			step = name_end_match
		#	
		#	else :
		#		#print("\nNewer names")
		#		step = name_end_match + 4
		#	
		#	gpu_list = ""
		#	gpu_char = t_efi_dump[step:step + 1]
		#	
		#	while True :
		#		gpu_name = ""
		#		
		#		while gpu_char != b'\x00' :
		#			gpu_name += gpu_char.decode('utf-8', 'ignore')
		#			step     += 1
		#			gpu_char = t_efi_dump[step:step + 1]
		#		
		#		if gpu_name[:11] in ['[ATI', 'Compromised', 'Protocol Er', 'Destination', 'ASSERT %a(%'] or len(gpu_name) < 3:
		#			break
		#		
		#		gpu_list += gpu_name + "\n"
		#						
		#		while gpu_char == b'\x00' :
		#			step     += 1
		#			gpu_char = t_efi_dump[step:step + 1]
		#		
		#	with open("%s_temp\\AMD_GOP_%s_names.txt" % (file_rom, t_version), "a") as myfile :
		#			myfile.write(gpu_list)
			
		return (t_gop_type, "", t_version, t_efi_info_string)

	## Nvidia GOP
	pat_nv = re.compile(br'\x4E\x56\x2D\x55\x45\x46\x49\x2D\x42\x4C\x44\x2D\x49\x4E\x46\x4F') ## NV-UEFI-BLD-INFO
	match_nv = pat_nv.search(t_efi_dump)
	
	if match_nv is not None :
		(nv_start_match, nv_end_match) = match_nv.span()
		t_gop_type = "Nvidia"
		date      = t_efi_dump[nv_end_match + 8:nv_end_match + 0x13].decode('utf-8', 'ignore')
		t_version = t_efi_dump[nv_end_match + 0x18:nv_end_match + 0x1F].decode('utf-8', 'ignore')
		chg       = t_efi_dump[nv_end_match + 0x24:nv_end_match + 0x2C].decode('utf-8', 'ignore')
		var_test  = t_efi_dump[nv_end_match + 0x2D:nv_end_match + 0x31]
		
		if var_test == b'VAR:' : ## 5641523A
			id_test = t_efi_dump[nv_end_match + 0x31:nv_end_match + 0x3C]
			
			if id_test == b'VARIANT_ID=' :
				var = t_efi_dump[nv_end_match + 0x3C:nv_end_match + 0x4C].decode('utf-8', 'ignore')
			else :
				var = t_efi_dump[nv_end_match + 0x31:nv_end_match + 0x41].decode('utf-8', 'ignore')
			
			if var[-2:] == "01" :
				t_nv_type = "GT21x"
			elif var[-2:] == "02" :
				t_nv_type = "GF10x"
			elif var[-2:] == "03" :
				t_nv_type = "GF119"
			elif var[-2:] == "04" :
				t_nv_type = "GK1xx"
			elif var[-2:] == "05" :
				t_nv_type = "GM1xx"
			elif var[-2:] == "06" :
				t_nv_type = "GM2xx"
			elif var[-2:] == "07" :
				t_nv_type = "GP1xx"
			else :
				t_nv_type = "GXnew"
			
			if var[:12] != "000000000000" :
				t_nv_type += "_Strange"
			
			if var[12:14] == "01" :
				t_nv_type += "_MXM"
			elif var[12:14] == "02" :
				t_nv_type += "_Multi-Display"
			elif var[12:14] == "03" :
				t_nv_type += "_GOP-Override"
			elif var[12:14] != "00" :
				t_nv_type += "_Custom"
						
			var = "0x" + var + " = " + t_nv_type
		else :
			if re.search(br'\x4D\x58\x4D\x5F', t_efi_dump) is not None : ## MXM_ or MXM_V (\x56|\x76)
				t_nv_type = "GXxxx_MXM"
				var = "missing = GXxxx_MXM"
			else :
				t_nv_type = "GXxxx"
				var = "missing"
			
			#var = "missing" if t_nv_type == "GXxxx" else "missing = GXxxx_MXM"
		
		print(Style.BRIGHT + Fore.GREEN + "Nvidia GOP" + Fore.WHITE + " %s  " % t_version + 
		Fore.GREEN + "Variant" + Fore.WHITE + " %s \n\n" % var + 
		Fore.GREEN + "Dated:" + Fore.WHITE + " %s  " % date + 
		Fore.GREEN + "Changelist" + Fore.WHITE + " %s" % chg + 
		Fore.RESET + Style.NORMAL)
		
		t_efi_info_string = t_nv_type + " - " + t_version + " - " + date + " - " + chg
		
		return (t_gop_type, t_nv_type, t_version, t_efi_info_string)
	
	## Nvidia GOP older versions
	
	pat_nv = re.compile(br'\xD6\xDC\x9B\xE1\xDF\xA6\xE4\x4F\xB7\x53\xD6\x77\xC7\x24\x8B\x77\x70\x21\x0A\xCC\x39\x0B\xD8\x45\xB4\x69\xD0\x4E\x57\xB5\x22\xB3\xEE\x27\xE5\xF5\x83\xA2\x27\x41\x9A\x2F\xAF\xFA\xBF\x26\xDF\x70') ## Part of a hash or something
	match_nv = pat_nv.search(t_efi_dump)
	
	if match_nv is not None :
		
		(nv_start_match, nv_end_match) = match_nv.span()
		t_gop_type = "Nvidia"
		
		test_ver1  = int.from_bytes(t_efi_dump[nv_start_match - 0x18:nv_start_match - 0x14], 'little')
		test_ver2  = int.from_bytes(t_efi_dump[nv_start_match + 0x108:nv_start_match + 0x10C], 'little')
		
		if 0x10000 <= test_ver2 <= 0x100017 :
			t_version = "0x%0.5X" % test_ver2
		elif 0x10000 <= test_ver1 <= 0x100017 :
			t_version = "0x%0.5X" % test_ver1
		else :
			t_version = "unknown"
		
		if re.search(br'\x4D\x58\x4D\x5F', t_efi_dump) is not None : ## MXM_ or MXM_V (\x56|\x76)
			t_nv_type = "GXxxx_MXM"
			var = "missing = GXxxx_MXM"
		else :
			t_nv_type = "GXxxx"
			var = "missing"
		
		date = "missing"
		chg  = "missing"
		
		print(Style.BRIGHT + Fore.GREEN + "Nvidia GOP" + Fore.WHITE + " %s  " % t_version + 
		Fore.GREEN + "Variant" + Fore.WHITE + " %s \n\n" % var + 
		Fore.GREEN + "Dated:" + Fore.WHITE + " %s  " % date + 
		Fore.GREEN + "Changelist" + Fore.WHITE + " %s" % chg + 
		Fore.RESET + Style.NORMAL)
		
		t_efi_info_string = t_nv_type + " - " + t_version + " - " + date + " - " + chg
		
		return (t_gop_type, t_nv_type, t_version, t_efi_info_string)
	
	## Mac AMD
	pat_mac_ati = re.compile(br'\x41\x00\x54\x00\x49\x00\x20\x00\x52\x00\x61\x00\x64\x00\x65\x00\x6F\x00\x6E\x00\x20\x00\x55\x00\x47\x00\x41\x00\x20\x00\x44\x00\x72\x00\x69\x00\x76\x00\x65\x00\x72\x00') ## A.T.I. .R.a.d.e.o.n. .U.G.A. .D.r.i.v.e.r.
	pat_mac_amd = re.compile(br'\x41\x00\x4D\x00\x44\x00\x20\x00\x52\x00\x61\x00\x64\x00\x65\x00\x6F\x00\x6E\x00\x20\x00\x44\x00\x72\x00\x69\x00\x76\x00\x65\x00\x72\x00')
	# A.M.D. .R.a.d.e.o.n. .D.r.i.v.e.r.
	
	match_mac_amd = pat_mac_ati.search(t_efi_dump)
	
	if match_mac_amd is None :
		match_mac_amd = pat_mac_amd.search(t_efi_dump)
		
	if match_mac_amd is not None :
		(mac_start_match, mac_end_match) = match_mac_amd.span()
		t_gop_type = "Mac_AMD"
		t_version  = t_efi_dump[mac_end_match + 2:mac_end_match + 0x14].decode('utf-16', 'ignore')
		
		pat_mac_date = re.compile(br'\x45\x00\x46\x00\x49\x00\x43\x00\x6F\x00\x6D\x00\x70\x00\x69\x00\x6C\x00\x65\x00\x44\x00\x61\x00\x74\x00\x65\x00')
		match_mac_date = pat_mac_date.search(t_efi_dump)
		
		if match_mac_date is not None :
			(date_start_match, date_end_match) = match_mac_date.span()
			date = t_efi_dump[date_start_match - 0x18:date_start_match - 0xD].decode('utf-8', 'ignore')
			
			if date[:3].isnumeric() :
				date = t_efi_dump[date_end_match + 4:date_end_match + 0xF].decode('utf-8', 'ignore')
		
		else :
			date = "not found"
		
		print(Style.BRIGHT + Fore.RED + "Mac AMD GOP" + Fore.WHITE + " %s  " % t_version + Fore.RED + "Dated:" + 
		Fore.WHITE + " %s  " % date + Fore.RESET + Style.NORMAL)
		
		pat_mac_ver = re.compile(br'\x41(\x54|\x4D)(\x49|\x44)\x20\x52\x61\x64\x65\x6F\x6E\x20\x48\x44\x20') # ATI|AMD Radeon HD
		match_mac_ver = pat_mac_ver.search(t_efi_dump)
		
		if match_mac_ver is not None :
			(mac_start_ver, mac_end_ver) = match_mac_ver.span()
			ver_list = " ## "
			step     = mac_start_ver
			ver_char = t_efi_dump[step:step + 1]
			
			while True :
				ver_item = ""
				
				while ver_char != b'\x00' :
					ver_item += ver_char.decode('utf-8', 'ignore')
					step     += 1
					ver_char = t_efi_dump[step:step + 1]
				
				if len(ver_item) < 3:
					break
				
				ver_list += ver_item + " ## "
								
				while ver_char == b'\x00' :
					step     += 1
					ver_char = t_efi_dump[step:step + 1]
		
		print(Style.BRIGHT + Fore.RED + "\nDetails:" + Fore.WHITE + "%s" % ver_list + Fore.RESET + Style.NORMAL)	
		
		return (t_gop_type, "", t_version, "")
	
	## Mac Nvidia
	
	#pat_mac_nv = re.compile(br'\x4E\x00\x56\x00\x49\x00\x44\x00\x49\x00\x41\x00\x20\x00\x47\x00\x50\x00\x55\x00\x20\x00')
	##\x55\x00\x45\x00\x46\x00\x49\x00\x20\x00\x44\x00\x72\x00\x69\x00\x76\x00\x65\x00\x72\x00') ## N.V.I.D.I.A. .G.P.U. .U.E.F.I. .D.r.i.v.e.r. - False positives with normal Nvidia GOP
	#match_mac_nv = pat_mac_nv.search(t_efi_dump)
	
	# NVDA,NVMac
	# AAPL,EMC-Display-List
	# AAPL,boot-display
	# AAPL,backlight-control
	# APPLE
	# APPLE_SRC_DPCD_ACCESS
	# MacVidCards
	
	if re.search(b'NVDA,NVMac', t_efi_dump) is not None :
		match_mac_nv = True
	elif re.search(br'\x41\x00\x41\x00\x50\x00\x4C\x00\x2C\x00', t_efi_dump) is not None : # A.A.P.L.,.
		match_mac_nv = True
	elif re.search(br'\x41\x00\x50\x00\x50\x00\x4C\x00\x45\x00', t_efi_dump) is not None : # A.P.P.L.E.
		match_mac_nv = True
	else :
		match_mac_nv = False
	
	#if match_mac_nv is not None :
		#(mac_start_match, mac_end_match) = match_mac_nv.span()
	if match_mac_nv :
		t_gop_type = "Mac_Nvidia"
		pat_mac_ver = re.compile(br'\x4E\x56\x44\x41\x2D\x45\x46\x49\x2D\x42\x75\x69\x6C\x64\x2D\x49\x6E\x66\x6F')
		match_mac_ver = pat_mac_ver.search(t_efi_dump)
		
		if match_mac_ver is not None :
			(mac_start_ver, mac_end_ver) = match_mac_ver.span()
			t_version = t_efi_dump[mac_end_ver:mac_end_ver + 0x13].decode('utf-8', 'ignore').replace("-", " ").strip()
		else :
			t_version = "unknown"
		
		if re.search(br'\x4D\x58\x4D\x5F', t_efi_dump) is not None : ## MXM_
			t_version += "_MXM"
		
		print(Style.BRIGHT + Fore.GREEN + "Mac Nvidia GOP" + Fore.WHITE + " %s  " % t_version + Fore.RESET + Style.NORMAL)
		
		return (t_gop_type, "", t_version, "")
	
	
	## LSI MPT UEFI
	## Added only for extraction purpose
	pat_lsi   = re.compile(br'\x4C\x53\x49\x20\x53\x41\x53(\x32|\x33)\x20\x4D\x50\x54\x20\x55\x45\x46\x49') ## LSI SASx MPT UEFI
	match_lsi = pat_lsi.search(t_efi_dump)
	
	if match_lsi is not None :
		t_gop_type = "LSI SASx MPT UEFI"
		(lsi_start_match, lsi_end_match) = match_lsi.span()
		sas_type  = t_efi_dump[lsi_start_match + 7:lsi_start_match + 8].decode('utf-8', 'ignore')
		t_gop_type = "LSI SAS%s MPT UEFI" % sas_type
		pat_ver   = re.compile(br'\x4C\x53\x49\x20\x43\x6F\x72\x70\x6F\x72\x61\x74\x69\x6F\x6E\x0A\x76') ## LSI Corporation.v
		match_ver = pat_ver.search(t_efi_dump, lsi_end_match)
		
		if match_ver is None :
			pat_ver   = re.compile(br'\x41\x76\x61\x67\x6F\x20\x54\x65\x63\x68\x6E\x6F\x6C\x6F\x67\x69\x65\x73\x2E\x20\x41\x6C\x6C\x20\x72\x69\x67\x68\x74\x73\x20\x72\x65\x73\x65\x72\x76\x65\x64\x2E\x0A\x76') ## Avago Technologies. All rights reserved..v
			match_ver = pat_ver.search(t_efi_dump, lsi_end_match)
		
		if match_ver is not None :
			(ver_start_match, ver_end_match) = match_ver.span()
			step = ver_end_match
			
			while t_efi_dump[step:step + 1] != b'\x0A' :
				step += 1
			
			t_version = t_efi_dump[ver_end_match:step].decode('utf-8', 'ignore').strip()
			print("LSI SAS%s MPT UEFI %s" % (sas_type, t_version))
			
			try :
				cut_ver   = t_version.index("*")
				t_version = t_version[:cut_ver].strip()
			except :
				pass
			
		return (t_gop_type, "", t_version, "")
	
	print(Style.BRIGHT + Fore.CYAN + "\nEFI ROM is present, but is not standard GOP type." + Fore.RESET + Style.NORMAL)
	
	return ("Unknown", "", "unknown", "")

def nvidia_board() :
	hint_message = "Missing"
	
	pat_nv   = re.compile(br'\x4E\x56\x49\x44\x49\x41\x20\x43\x6F\x72\x70\x2E\x0D\x0A') ## NVIDIA Corp...
	match_nv = pat_nv.search(reading)
	
	if match_nv is not None :
		(nv_start_match, nv_end_match) = match_nv.span()
		
		if reading[nv_end_match:nv_end_match + 0xB] == b'\x00\x00\x00\xFF\xFF\x00\x00\x00\x00\xFF\xFF' :
			#print("Normal string")
			hint_bgn = nv_end_match + 0xB
		else :
			#print("Different string")
			hint_bgn  = nv_end_match
			skip_byte = reading[hint_bgn:hint_bgn + 1]
			
			while skip_byte == b'\x00' or skip_byte == b'\xFF' :
				hint_bgn  += 1
				skip_byte = reading[hint_bgn:hint_bgn + 1]
		
		hint_message = ""
		hint_step    = hint_bgn
		hint_char    = reading[hint_step:hint_step + 1]
		
		while hint_char != b'\x00' and hint_char != b'\xFF' and hint_char != b'\x2D' :
			hint_message += hint_char.decode('utf-8', 'ignore')
			hint_step    += 1
			hint_char    = reading[hint_step:hint_step + 1]
	
	return hint_message.strip()

def isbn_struct(nv_image, print_info) :
	isbn_sig   = nv_image[:4].decode('utf-8', 'ignore')
	isbn_hdr   = int.from_bytes(nv_image[4:8], 'little')
	flag_v1    = ord(nv_image[8:9])
	flag_v2    = ord(nv_image[9:0xA])
	bv_sig_v1  = nv_image[0xA:0xC].decode('utf-8', 'ignore')
	checksm_v1 = int.from_bytes(nv_image[0xC:0x10], 'little')
	checksm_v2 = int.from_bytes(nv_image[0x10:0x14], 'little')
	bv_sig_v2  = nv_image[0x14:0x16].decode('utf-8', 'ignore')
	isbn_size  = int.from_bytes(nv_image[0x16:0x18], 'little') # maybe 4 bytes
	flag_v3    = int.from_bytes(nv_image[0x18:0x1C], 'little')
	
	if print_info :
		print("----------ISBN_HEADER---------\n")
		print("Signature:                      %s"      % isbn_sig)
		print("Header Size:                    0x%0.2X" % isbn_hdr)
		print("Unk01:                          0x%0.2X" % flag_v1)
		print("Unk02:                          0x%0.2X" % flag_v2)
		print("BV Signature:                   %s"      % bv_sig_v1)
		print("Checksum v1:                    0x%0.2X" % checksm_v1)
		print("Checksum v2:                    0x%0.2X" % checksm_v2)
		print("BV Signature:                   %s"      % bv_sig_v2)
		print("ISBN Body Size:                 0x%0.2X" % isbn_size)
		print("Unk03:                          0x%0.2X" % flag_v3)
		print("\n------------------------------\n")
	
	isbn_step = 0x1C # Hard coded ?
	isbn_end  = isbn_hdr + isbn_size
	count     = 0
	old_size  = 0
	
	while isbn_step < isbn_end :
		
		# First is a certificate
		# Second is a certificate
		# Third is a map
		# Fourth is a hash
		
		count     += 1
		
		cert_type = int.from_bytes(nv_image[isbn_step:isbn_step + 4], 'little')
		cert_hdr  = int.from_bytes(nv_image[isbn_step + 4:isbn_step + 8], 'little')
		map_el_nr = int.from_bytes(nv_image[isbn_step + 8:isbn_step + 0xC], 'little')
		cert_size = int.from_bytes(nv_image[isbn_step + 0xC:isbn_step + 0x10], 'little')
		cert_flg1 = int.from_bytes(nv_image[isbn_step + 0x10:isbn_step + 0x14], 'little')
		cert_flg2 = int.from_bytes(nv_image[isbn_step + 0x14:isbn_step + 0x18], 'little')
		cert_flg3 = int.from_bytes(nv_image[isbn_step + 0x18:isbn_step + 0x1C], 'little')
		cert_flg4 = int.from_bytes(nv_image[isbn_step + 0x1C:isbn_step + 0x20], 'little')
		cert_flg5 = int.from_bytes(nv_image[isbn_step + 0x20:isbn_step + 0x24], 'little')
		
		cert_bgn  = isbn_step + cert_hdr
		
		if print_info :
			
			print("----------CERT_HEADER---------\n")
			print("Type:                           0x%0.2X" % cert_type)
			print("Header Size:                    0x%0.2X" % cert_hdr)
			print("Nr of map elements:             0x%0.2X" % map_el_nr)
			print("Certificate size:               0x%0.2X" % cert_size)
			
			if cert_hdr == 0x24 :
				
				date_str = datetime.datetime.fromtimestamp(cert_flg3).strftime('%Y-%m-%d %H:%M:%S')
				
				print("Flags v1 - Sign Size:           0x%0.2X" % cert_flg1)
				print("Flags v2 - Hash Size:           0x%0.2X" % cert_flg2)
				print("Flags v3 - Date:                0x%0.2X" % cert_flg3 + " = " + date_str)
				print("Flags v4:                       0x%0.2X" % cert_flg4)
				print("Flags v5:                       0x%0.2X" % cert_flg5)
			
			elif cert_hdr == 0x1C :
				
				print("Flags v1 - Cert Type:           0x%0.2X" % cert_flg1)
				print("Flags v2 - Issuer ID:           0x%0.2X" % cert_flg2)
				print("Flags v3 - Owner ID:            0x%0.2X" % cert_flg3)
			
			elif cert_hdr > 0x10 :
				flag_nr   = 0
				flag_step = isbn_step + 0x10
				
				for ix in range(0, cert_hdr - 0x10, 4) :
					flag_nr   += 1
					cert_flgx = int.from_bytes(nv_image[flag_step:flag_step + 4], 'little')
					
					print("Flags v%0.2d :                     0x%0.2X" % (flag_nr, cert_flgx))
					
					flag_step += 4
			
			print("\n------------------------------\n")
		
		if count == 2 :
			cert_size -= old_size
		else :
			old_size  = cert_size
		
		if map_el_nr < 0x100 and count == 3 :
			
			map_step = cert_bgn
			
			for ix in range(map_el_nr) :
				
				if ix == 0 and print_info :
					print("----------MAP_Content---------\n")
					print(" Offset  | Size    | Flags \n")
				
				elem_off  = int.from_bytes(nv_image[map_step:map_step + 4], 'little')
				elem_size = int.from_bytes(nv_image[map_step + 4:map_step + 8], 'little')
				elem_flag = int.from_bytes(nv_image[map_step + 8:map_step + 0xC], 'little')
				map_step  += 0xC
				
				if print_info :
					print(" %-8X  %-8X  %-8X" % (elem_off, elem_size, elem_flag))
			
			if print_info :
				print("\n------------------------------\n")
		
		if nv_image[cert_bgn:cert_bgn + 11] == b'-----BEGIN ' :
			
			if print_info :
				print(nv_image[cert_bgn:isbn_step + cert_size].decode('utf-8', 'ignore'))
			
			with open(file_rom + "_temp\\cert_nr%d.cer" % count, 'wb') as cert_file :
				cert_file.write(nv_image[cert_bgn:isbn_step + cert_size])
			
			print(Style.BRIGHT + Fore.YELLOW + "Extracted cert_nr%d.cer\n" % count + Fore.RESET + Style.NORMAL)
		
		elif map_el_nr == 0 :
			
			with open(file_rom + "_temp\\cert_nr%d.lic" % count, 'wb') as cert_file :
				cert_file.write(nv_image[cert_bgn:isbn_step + cert_size])
			
			print(Style.BRIGHT + Fore.YELLOW + "Extracted cert_nr%d.lic\n" % count + Fore.RESET + Style.NORMAL)
		
		# else :
			
			# with open(file_rom + "_temp\\cert_nr%d.unk" % count, 'wb') as cert_file :
				# cert_file.write(nv_image[cert_bgn:isbn_step + cert_size])
			
			# print(Style.BRIGHT + Fore.YELLOW + "Extracted cert_nr%d.unk\n" % count + Fore.RESET + Style.NORMAL)
		
		#print(count)
		
		if cert_size == 1 or count == 4 :
			
			cert_size = cert_hdr + cert_flg1 + cert_flg2
			hash_v1   = binascii.hexlify(nv_image[cert_bgn:cert_bgn + 0x80]).decode('utf-8').upper()
			hash_v2   = binascii.hexlify(nv_image[cert_bgn + 0x80:cert_bgn + 0x90]).decode('utf-8').upper()
			
			if print_info :
				print("---------SIGN_Content---------\n")
				print("\n".join(hash_v1[xstr:xstr + 32] for xstr in range(0, len(hash_v1), 32)))
				print("\n------------------------------\n")
				print("----------Image_HASH----------\n")
				print(hash_v2)
				print("\n------------------------------\n")
		
		if cert_type > 0x10 or count == 5 :
			cert_size = int.from_bytes(nv_image[isbn_step + 0x14:isbn_step + 0x18], 'little')
			flag_nr   = 0
			flag_step = isbn_step + cert_hdr
			
			if print_info :
				print("----------MAP_Content---------\n")
				print(" Flags   | Offset  | Size    | Hash\n")
			
			for ix in range(0, cert_size - cert_hdr - 0x10, 0x1C) :
				flag_nr   += 1
				cert_flgx = int.from_bytes(nv_image[flag_step:flag_step + 4], 'little')
				cert_flgy = int.from_bytes(nv_image[flag_step + 4:flag_step + 8], 'little')
				cert_flgz = int.from_bytes(nv_image[flag_step + 8:flag_step + 0xC], 'little')
				cert_hash = binascii.hexlify(nv_image[flag_step + 0xC:flag_step + 0x1C]).decode('utf-8').upper()
				
				if print_info :
					# print("Flags :                         0x%0.2X" % cert_flgx)
					# print("Offset :                        0x%0.2X" % cert_flgy)
					# print("Size :                          0x%0.2X" % cert_flgz)
					# print("Hash :                          %s"      % cert_hash)
					print(" %-8X  %-8X  %-8X  %s" % (cert_flgx, cert_flgy, cert_flgz, cert_hash))
					print()
				
				flag_step += 0x1C
			
			mast_hash = binascii.hexlify(nv_image[isbn_step + cert_size - 0x10:isbn_step + cert_size]).decode('utf-8').upper()
			
			if print_info :
				print("Master Hash :                  %s" % mast_hash)
				print("\n------------------------------\n")
		
		if count == 6 :
			cert_size = isbn_end - isbn_step
			hash_v1   = binascii.hexlify(nv_image[cert_bgn:isbn_step + cert_size]).decode('utf-8').upper()
			
			if print_info :
				print("---------SIGN_Content---------\n")
				print("\n".join(hash_v1[xstr:xstr + 32] for xstr in range(0, len(hash_v1), 32)))
				print("\n------------------------------\n")
		
		if cert_size :
			isbn_step += cert_size
		else :
			isbn_step += cert_hdr + map_el_nr * 0xC

def is_version_upd(t_last_gop, t_version, t_gop_type) :
	
	if t_gop_type == "AMD" :
		t_lastgop_int = int(t_last_gop[2:4])
		t_version_int = int(t_version[2:4].strip("."))
	elif t_gop_type == "Nvidia" :
		t_lastgop_int = int(t_last_gop, 16)
		t_version_int = int(t_version, 16)
	
	if t_version == t_last_gop :
		print("You already have the latest available GOP!\n")
		sys.exit()
	elif t_version_int < t_lastgop_int :
		print("Latest available GOP is %s\n" % t_last_gop)
	else :
		print(Style.BRIGHT + Fore.YELLOW + "You have a newer version! Please report it in the forum!\n" + Fore.RESET + Style.NORMAL)
		sys.exit()
	
	return None

def remove_padding(t_end_data, t_end_img_old, t_end_img_new, t_all_size, t_str_err) :
	bgn_extra = 0
	
	for padd_off in range (t_end_img_old, t_all_size) :
		padd_byte = reading[padd_off:padd_off + 1]
		
		if padd_byte == b'' :
			break
		
		elif padd_byte != b'\xFF' and padd_byte != b'\x00' :
			bgn_extra = padd_off
			print(Style.BRIGHT + Fore.RED + t_str_err + Fore.RESET)
			#t_end_data = reading[t_end_img_old:]
			break
	
	## If there is extra data after ROM images.
	if bgn_extra :
		## If the extra data offset is now inside the new image.
		
		if bgn_extra < t_end_img_new :
			print(Style.BRIGHT + Fore.RED + "  Unable to recover extra data at the same offset 0x%0.2X! Please report it!\n" % bgn_extra + 
			Fore.RESET + Style.NORMAL)
			t_end_data = reading[t_end_img_old:]
		
		## Extra data can be recovered at the old offset
		else :
			print(Style.BRIGHT + Fore.YELLOW + "  Recovering extra data at the same offset 0x%0.2X.\n" % bgn_extra + 
			Fore.RESET + Style.NORMAL)
			## If the new image ends after the old one, just copy from there.
			
			if t_end_img_new > t_end_img_old :
				t_end_data = reading[t_end_img_new:]
			
			## But if the old image was bigger, we need to fill the difference with padding and then copy from the end of old image.
			else :
				t_end_data = b'\xFF' * (t_end_img_old - t_end_img_new) + reading[t_end_img_old:]
	
	return t_end_data

def check_in_database(t_efi_info_string) :
	efi_in_db     = False
	bad_nvd       = False
	bad_amd       = False
	#unknown_type = False
	
	#if t_efi_info_string == "" :
	#	return None
	
	if nv_type == "GXxxx" or nv_type == "GXxxx_MXM" :
		unknown_type = True
		index        = len(nv_type) + 3
	else :
		unknown_type = False
		index        = 0
	
	with open("#GOP_Files\\#GOP_Database.txt", 'r+') as db_file:
		
		for line in db_file:
			
			if len(line) < 2 :
				continue
			
			elif line[:2] == "##" :
				
				if line[3:13] == "BAD_NVIDIA" :
					bad_nvd = True
				elif line[3:10] == "BAD_AMD" :
					bad_amd = True
				else :
					continue
			
			if line[index:] == t_efi_info_string[index:] :
				
				efi_in_db = True
				
				if gop_type == "AMD" and bad_amd :
					print(Style.BRIGHT + Fore.YELLOW + "You have a broken EFI image!\n" + Fore.RESET + Style.NORMAL)
				elif gop_type == "Nvidia" and bad_nvd :
					print(Style.BRIGHT + Fore.YELLOW + "You have a broken EFI image!\n" + Fore.RESET + Style.NORMAL)
				
				if unknown_type :
					new_type = line[:index - 3]
				
				#print("EFI %s is present in the database!\n" % t_efi_info_string)
				break
	
	if efi_in_db :
		
		if unknown_type :
			return new_type, True
		else :
			return None, True

	else :
		print(Style.BRIGHT + Fore.YELLOW + "Note: The GOP file is not present in my database.\n\n      You can help me by reporting it.\n" + 
		Fore.RESET + Style.NORMAL)
		
		return None, False

####################################
####################################
####################################

if len(sys.argv) < 3 :
	print(Fore.RED + "Not enough arguments! Usage: GOPupd.py file.rom [ext_efirom | gop_upd | isbn]" + Fore.RESET)
	sys.exit()
else :
	file_dir   = sys.argv[1]
	file_rom   = os.path.basename(file_dir)
	file_arg   = sys.argv[2]
	extra_args = sys.argv[3:]

if not os.path.isfile(file_dir) :
	print(Fore.RED + "File %s was not found!" % file_dir + Fore.RESET)
	sys.exit()

try :
	f = open(file_dir, 'rb')
	reading = f.read()
	f.close()
except :
	print(Fore.RED + "Unable to open file %s for reading!" % file_dir + Fore.RESET)
	sys.exit()

try :
	os.mkdir(file_rom + "_temp\\")
except :
	pass

fileName, fileExtension = os.path.splitext(file_rom)
position = 0

if file_arg == "ext_efirom" :
	## Get ROM info for EFI extraction
	efi_found, efi_begin, efi_size = rom_info(reading, "mini")
	
	if not efi_found :
		print(Fore.RED + "No EFI ROM found!\n" + Fore.RESET)
		#f.close()
		sys.exit()
	
	efi_rom  = reading[efi_begin:efi_begin + efi_size]
	
	with open("%s_temp\\%s_compr.efirom" % (file_rom, fileName), 'wb') as efi_rom_file :
		efi_rom_file.write(efi_rom)

elif file_arg.upper() == "ISBN" :
	rom_old_type, rom_start, rom_pcir_off, rom_id_bin, rom_size, efi_found, efi_begin, efi_size = rom_info(reading, "all")
	
	if "info" in extra_args :
		print_info = True
	else :
		print_info = False
	
	isbn_found = False
	nv_step    = rom_start + rom_size
	
	if reading[rom_pcir_off + 0x20:rom_pcir_off + 0x24] == b'NPDE' :
		npde_size = int.from_bytes(reading[rom_pcir_off + 0x28:rom_pcir_off + 0x2A], 'little') * 0x200
		
		if rom_size != npde_size :
			print("  Different sizes in PCI structure and NPDE structure of Legacy ROM!\n")
			
			rom_end_npde  = rom_start + npde_size
			rom_test_npde = reading[rom_end_npde:rom_end_npde + 2]
			
			## If NPDE size is the right one, there is one container for all ROMs.
			if rom_test_npde in [b'\x55\AA', b'\x56\x4E'] :
				print("  The Legacy ROM appears to be a container for all images.\n")
				nv_step      = rom_end_npde
			
	check_nv_ext = reading[nv_step:nv_step + 2]
				
	while check_nv_ext in [b'\x56\x4E', b'\x55\xAA', b'\x77\xBB'] :
		
		if check_nv_ext == b'\x77\xBB' :
			npds_off    = int.from_bytes(reading[nv_step + 0x18:nv_step + 0x1A], 'little')
			npds_off    += nv_step
			nv_ext_size = int.from_bytes(reading[npds_off + 0x10:npds_off + 0x12], 'little') * 0x200
		
		else :
			nv_ext_size = int.from_bytes(reading[nv_step + 2:nv_step + 4], 'little') * 0x200
		
		#print(nv_ext_size)
		
		if nv_ext_size == 0 :
			npds_off    = int.from_bytes(reading[nv_step + 0x18:nv_step + 0x1A], 'little')
			npds_off    += nv_step
			nv_ext_size = int.from_bytes(reading[npds_off + 0x10:npds_off + 0x12], 'little') * 0x200
			#print(nv_ext_size)
			
			npds_str_size = int.from_bytes(reading[npds_off + 0xA:npds_off + 0xC], 'little')
			isbn_off      = npds_off + npds_str_size
			isbn_check    = reading[isbn_off:isbn_off + 4]
			
			if isbn_check == b'ISBN' :
				print(Style.BRIGHT + Fore.YELLOW + "Found ISBN at offset 0x%0.2X\n" % isbn_off + Fore.RESET + Style.NORMAL)
				isbn_struct(reading[isbn_off:nv_step + nv_ext_size], print_info)
				isbn_found = True
		
		nv_step      += nv_ext_size
		check_nv_ext = reading[nv_step:nv_step + 2]
	
	if not isbn_found :
		print(Style.BRIGHT + Fore.YELLOW + "ISBN was not found!\n" + Fore.RESET + Style.NORMAL)

elif file_arg == "gop_upd" :

	gop_type = ""
	file_efi = "%s_temp\\%s_dump.efi" % (file_rom, fileName)
	file_efr = "%s_temp\\%s_compr.efirom" % (file_rom, fileName)
	
	if os.path.isfile(file_efi) :
		with open(file_efi, 'rb') as myfile :
			efi_dump = myfile.read()
		
		## Get EFI info
		gop_type, nv_type, version, efi_info_string = efi_version(efi_dump)
		
		## The machine code type, signer and CRC32 are processed outside efi_version because they are not bound to GOP.
		
		## Signer
		## Needs a full X.509 parser. Since it is not that important, only get the first signer.
		pe_off     = int.from_bytes(efi_dump[0x3C:0x40], 'little')
		code_size  = int.from_bytes(efi_dump[pe_off + 0x50:pe_off + 0x54], 'little') ## only works for EFI files, not for every exe.
		pat_sign   = re.compile(br'\x06\x09\x2A\x86\x48\x86\xF7\x0D\x01\x07\x02')
		match_sign = pat_sign.search(efi_dump, code_size)
		
		if match_sign is not None :
			(sign_start_match, sign_end_match) = match_sign.span()
			#print("\nEFI Image is signed!")
			efi_is_signed = "Signed"
			pat_signer    = re.compile(br'\x06\x03\x55\x04\x03\x13') ## 06 03 55 04 03 13
			match_signer  = pat_signer.search(efi_dump, sign_end_match)
			
			if match_signer is not None :
				(signer_start_match, signer_end_match) = match_signer.span()
				mess_len = ord(efi_dump[signer_end_match:signer_end_match + 1])
				signer   = efi_dump[signer_end_match + 1:signer_end_match + mess_len + 1].decode('utf-8', 'ignore')
				print(Style.BRIGHT + Fore.CYAN + "\nMost likely signed by: %s\n" % signer + Fore.RESET + Style.NORMAL)
			else :
				print(Style.BRIGHT + Fore.CYAN + "\nEFI Image is signed!\n" + Fore.RESET + Style.NORMAL)
		
		else :
			efi_is_signed = "Unsigned"
			print(Style.BRIGHT + Fore.CYAN + "\nEFI image is NOT signed!\n" + Fore.RESET + Style.NORMAL)
		
		if gop_type == "AMD" :
			efi_info_string += " - " + "%s" % efi_is_signed
		
		## Machine Code Type
		#pe_off  = int.from_bytes(efi_dump[0x3C:0x40], 'little')
		pe_test = efi_dump[pe_off:pe_off + 4]
		
		if pe_test == b'\x50\x45\x00\x00' :
			code = binascii.hexlify(efi_dump[pe_off + 4:pe_off + 6][::-1]).decode('utf-8').upper()
			
			if code == "014C" :
				codetype = "IA32"
			elif code == "0200" :
				codetype = "IA64"
			elif code == "8664" :
				codetype = "x64"
			elif code == "0EBC" :
				codetype = "EFI Byte Code"
			elif code == "01C0" :
				codetype = "ARM"
			elif code == "01C2" :
				codetype = "ARM-THUMB-MIXED"
			elif code == "01C4" :
				codetype = "ARMv7"
			else :
				codetype = "unknown (%s)" % code
			
			print(Style.BRIGHT + Fore.CYAN + "Machine Code   = %s\n" % codetype + Fore.RESET + Style.NORMAL)
		
		## CRC32
		#efi_crc32 = get_crc32(efi_dump)
		efi_crc32_int = binascii.crc32(efi_dump) & 0xFFFFFFFF
		efi_crc32_hex = "%08X" % efi_crc32_int
		print(Style.BRIGHT + Fore.CYAN + "Checksum CRC32 = %s\n" % efi_crc32_hex + Fore.RESET + Style.NORMAL)
		
		## Check in database
		if efi_info_string != "" :
						
			efi_info_string += " - " + "%s\n" % efi_crc32_hex
			
			#print(efi_info_string)
			#with open("#add_new_string.txt", "a") as myfile :
			#	myfile.write(efi_info_string)
			
			new_nv_type, efi_in_db = check_in_database(efi_info_string)
			
			if new_nv_type is not None :
				nv_type = new_nv_type
				#print("GOP identified as %s based on CRC\n" % nv_type)
				print(Style.BRIGHT + Fore.WHITE + "GOP identified as" + Fore.GREEN + " %s " % nv_type + 
				Fore.WHITE + "based on CRC \n" + Fore.RESET + Style.NORMAL)
		
		## Rename temp files
		
		extra_ver = ""
		pat_lg    = br'\x41(\x4D\x44|\x54\x49)\x20\x41\x54\x4F\x4D\x42\x49\x4F\x53\x00' ## AMD ATOMBIOS or ATI ATOMBIOS.
		
		if gop_type == "AMD" :
			
			# TODO Check for ROM, get CRC offset, compare to 0.
			# Usually a ROM in EFI means legacy tables are present.
			if (rom_info(efi_dump, "basic") or re.search(pat_lg, efi_dump) is not None) :
				extra_ver += "_custom_%s_%s" % (efi_is_signed.lower(), efi_crc32_hex)
			else :
				extra_ver += "_%s_%s" % (efi_is_signed.lower(), efi_crc32_hex)
		
		version_xt = version + extra_ver
		
		## For you
		
		gop_version  = "%s %s" % (nv_type, version_xt) if nv_type != "" else version_xt
		file_new_bgn = "%s_temp\\%s GOP %s" % (file_rom, gop_type, gop_version)
		
		## For me
		# version_me      = version_xt[2:] if version_xt[:2] == "0x" else version_xt
		# gop_version_me  = "%s %s" % (nv_type, version_me) if nv_type != "" else version_me
		# file_new_bgn    = "%s_temp\\%s GOP %s %s" % (file_rom, gop_type, gop_version_me, fileName)
		
		## And nothing for the rest.
		
		file_new_efr = "%s_compr.efirom" % file_new_bgn.rstrip()
		file_new_efi = "%s_dump.efi" % file_new_bgn.rstrip()
		
		if file_efi != file_new_efi and os.path.exists(file_new_efi) :
			os.remove(file_new_efi)
		if file_efr != file_new_efr and os.path.exists(file_new_efr) :
			os.remove(file_new_efr)
			
		try :
			os.rename(file_efi, file_new_efi)
		except Exception as e:
			print("Error on renaming temp files!\n")
			print(e)
			print()
		
		try :
			os.rename(file_efr, file_new_efr)
		except Exception as e:
			print("Error on renaming temp files!\n")
			print(e)
			print()
		
		if gop_type not in ['AMD', 'Nvidia'] or not efi_in_db :
			
			# if efi_info_string != "" :
				# with open("#add_new_string.txt", "a") as myfile :
					# myfile.write(efi_info_string)
			
			src_dir = "%s_temp" % file_rom
			dst_dir = "%s_newGOP" % file_rom
			os.rename(src_dir, dst_dir)
		
	print(Fore.RED + "---------------------------------------------------------------\n" + Fore.RESET)
	
	print(Fore.GREEN + "***************************************************************")
	print("***                Processing with Python...                ***")
	print("*************************************************************** \n" + Fore.RESET)
	#print("---------------------------------------------------------------\n\n")
	
	last_amd_new  = "1.63.0.15.50"
	last_amd_old  = "1.57.0.0.0"
	last_nv_GT21x = "0x10031"
	last_nv_GF10x = "0x1001F"
	last_nv_GF119 = "0x10030"
	last_nv_GK1xx = "0x10034"
	last_nv_GM1xx = "0x10036"
	last_nv_GM2xx = "0x20011"
	last_nv_GP1xx = "0x30008"
	
	last_nv_GK1xx_MXM = "0x10033"
	last_nv_GM1xx_MXM = "0x10034"
	last_nv_GK1xx_MDP = "0x10030"
	
	## GOP Test
	if gop_type == "AMD" :
		last_gop = last_amd_new
		is_version_upd(last_gop, version, gop_type)
	
	elif gop_type == "Nvidia" :
		if len(nv_type) > 12 and nv_type[6:13] == "Strange" : # GXxyz_Strange or GXxyz_Strange_[MXM|Multi-Display|Custom]
			print(Style.BRIGHT + Fore.YELLOW + "You have a strange GOP! Please report it!\n" + Fore.RESET + Style.NORMAL)
			#f.close()
			sys.exit()
		elif len(nv_type) > 6 and nv_type[-6:] == "Custom" : # GXxyz[_Strange]_Custom
			print(Style.BRIGHT + Fore.YELLOW + "You have a custom GOP! Currently not supported. Please report it!\n" + Fore.RESET + Style.NORMAL)
			#f.close()
			sys.exit()
		elif nv_type == "GT21x" :
			last_gop = last_nv_GT21x
			nv_file  = "nv_gop_GT21x.efirom"
			is_version_upd(last_gop, version, gop_type)
		elif nv_type == "GF10x" :
			last_gop = last_nv_GF10x
			nv_file  = "nv_gop_GF10x.efirom"
			is_version_upd(last_gop, version, gop_type)
		elif nv_type == "GF119" :
			last_gop = last_nv_GF119
			nv_file  = "nv_gop_GF119.efirom"
			is_version_upd(last_gop, version, gop_type)
		elif nv_type == "GK1xx" :
			last_gop = last_nv_GK1xx
			nv_file  = "nv_gop_GK1xx.efirom"
			is_version_upd(last_gop, version, gop_type)
		elif nv_type == "GM1xx" :
			last_gop = last_nv_GM1xx
			nv_file  = "nv_gop_GM1xx.efirom"
			is_version_upd(last_gop, version, gop_type)
		elif nv_type == "GM2xx" :
			last_gop = last_nv_GM2xx
			nv_file  = "nv_gop_GM2xx.efirom"
			is_version_upd(last_gop, version, gop_type)
		elif nv_type == "GP1xx" :
			last_gop = last_nv_GP1xx
			nv_file  = "nv_gop_GP1xx.efirom"
			is_version_upd(last_gop, version, gop_type)
		elif nv_type == "GK1xx_MXM" :
			last_gop = last_nv_GK1xx_MXM
			nv_file  = "nv_gop_GK1xx_MXM.efirom"
			is_version_upd(last_gop, version, gop_type)
		elif nv_type == "GM1xx_MXM" :
			last_gop = last_nv_GM1xx_MXM
			nv_file  = "nv_gop_GM1xx_MXM.efirom"
			is_version_upd(last_gop, version, gop_type)
		elif nv_type == "GK1xx_Multi-Display" :
			last_gop = last_nv_GK1xx_MDP
			nv_file  = "nv_gop_GK1xx_multi.efirom"
			is_version_upd(last_gop, version, gop_type)
		elif len(nv_type) > 3 and nv_type[-3:] == "MXM" : # GXxyz_MXM or GXxxx_MXM
			print(Style.BRIGHT + Fore.YELLOW + "You have an unsupported MXM GPU! Please report it! \n" + Fore.RESET + Style.NORMAL)
			#f.close()
			sys.exit()
		elif len(nv_type) > 13 and nv_type[-13:] == "Multi-Display" : # GXxyz_Multi-Display
			print(Style.BRIGHT + Fore.YELLOW + "You have an unsupported Multi-Display GPU! Please report it!\n" + Fore.RESET + Style.NORMAL)
			#f.close()
			sys.exit()
		elif nv_type != "GXxxx" : # GXnew
			print(Style.BRIGHT + Fore.YELLOW + "You have a new GOP type! Please report it!\n" + Fore.RESET + Style.NORMAL)
			#f.close()
			sys.exit()
		else : # == GXxxx i.e. no variant ID.
			last_gop = "latest available"
			print(Style.BRIGHT + Fore.YELLOW + "Unable to determine GOP type!\n" + Fore.RESET + Style.NORMAL)
	
	elif gop_type[:4] == "Mac_" :
		print(Style.BRIGHT + Fore.RED + "Mac GOP support is limited! Drop your compressed GOP as mac_gop.efirom in #GOP_Files\n" + 
		Fore.RESET + Style.NORMAL)
		last_gop = "your file"
		mac_file  = "mac_gop.efirom"
		#sys.exit()
	elif gop_type[:3] == "LSI" :
		print(Style.BRIGHT + Fore.RED + "LSI SASx MPT UEFI not supported!\n" + Fore.RESET + Style.NORMAL)
		sys.exit()
	elif gop_type == "Unknown" :
		print(Style.BRIGHT + Fore.RED + "Not GOP or GOP is not common type! Please report it!\n" + Fore.RESET + Style.NORMAL)
		sys.exit()
	else :
		last_gop = "latest available"
		print(Style.BRIGHT + Fore.RED + "GOP is not present!!!\n" + Fore.RESET + Style.NORMAL)
	
	## Get ROM Info
	orom_old_type, orom_start, orom_pcir_off, orom_id_bin, orom_size, efi_found, efi_begin, efi_size = rom_info(reading, "all")
	#ven_dev = "%s-%s" % (orom_id_hex[4:], orom_id_hex[:4])
	#pci_ven = ven_dev[:4]
	#pci_dev = ven_dev[-4:]
	ven_dev, pci_ven, pci_dev = id_from_bin(orom_id_bin, "all_list")
	efi_class_bin = reading[orom_pcir_off + 0xD:orom_pcir_off + 0x10] # Initialize for when GOP ROM missing
	#print(ven_dev)
	#print("%0.2X = %s-%s = %0.2X" % (orom_start, pci_ven, ven_dev[-4:], orom_last_img))
	#print(Style.BRIGHT + Fore.CYAN + "%s = ID of ROM file\n" % ven_dev + Fore.RESET + Style.NORMAL)
	
	
	## Check that EFI ROM header is a good match
	
	if orom_start == efi_begin :
		print(Style.BRIGHT + Fore.CYAN + "It appears you used an EFI ROM! Only extraction is possible." + Fore.RESET + Style.NORMAL)
		#f.close()
		sys.exit()
	elif pci_ven not in ['1002', '10DE'] :
		print(Style.BRIGHT + Fore.YELLOW + "Only AMD and Nvidia GOP supported!" + Fore.RESET + Style.NORMAL)
		#f.close()
		sys.exit()
	elif efi_found :
		if os.path.isfile(file_efi) and codetype != "x64" and gop_type[:4] != "Mac_" :
			print(Style.BRIGHT + Fore.YELLOW + "Code type %s is not supported!" % codetype + Fore.RESET + Style.NORMAL)
			#f.close()
			sys.exit()
		
		efi_pcir_ds    = int.from_bytes(reading[efi_begin + 0x18:efi_begin + 0x1A], 'little')
		efi_pcir_off   = efi_begin + efi_pcir_ds
		efi_class_bin  = reading[efi_pcir_off + 0xD:efi_pcir_off + 0x10]
		efi_class_code = binascii.hexlify(efi_class_bin).decode('utf-8')
		efi_code_type  = reading[efi_pcir_off + 0x14:efi_pcir_off + 0x15]
		efi_last_img   = ord(reading[efi_pcir_off + 0x15:efi_pcir_off + 0x16]) & 0x80
		
		if efi_class_code not in ['030000', '000003', '030200', '000203'] and gop_type[:4] != "Mac_" :
			print(Style.BRIGHT + Fore.YELLOW + "Class-code %s is not supported!" % efi_class_code + Fore.RESET + Style.NORMAL)
			#f.close()
			sys.exit()
		if efi_code_type != b'\x03' :
			print(Style.BRIGHT + Fore.RED + "Code mismatch in EFI ROM header!" + Fore.RESET + Style.NORMAL)
			#f.close()
			sys.exit()
		if not efi_last_img :
			print(Style.BRIGHT + Fore.RED + "EFI ROM is not last image! Please report it!" + Fore.RESET + Style.NORMAL)
			#f.close()
			sys.exit()
	
	## Pretty please
	ask = input("\nDo you want to update GOP to %s? Y for yes or N for no: " % last_gop)
	if ask.upper() != "Y" :
		#f.close()
		sys.exit()
	print("")
	
	## GRID K1/K2 check
	if pci_ven == "10DE" and pci_dev in ['0FF2', '11BF'] :
		gop_type = "Nvidia"
		nv_type  = "GK1xx_Multi-Display"
		nv_file  = "nv_gop_GK1xx_multi.efirom"
		last_gop = last_nv_GK1xx_MDP
		print(Style.BRIGHT + Fore.YELLOW + "  Using Multi-Display GOP %s for GRID K1/K2.\n" % last_nv_GK1xx_MDP + Fore.RESET + Style.NORMAL)
	
	## Get the right GOP
	if gop_type == "AMD" or (pci_ven == "1002" and gop_type == "") :
		## The next two lines are needed for the case of missing GOP.
		gop_type   = "AMD"
		last_gop   = last_amd_new
		efi_id_off = 0x20 ## might change it future versions
		id_in_gop  = False
		
		## GOP 1.59.0.0.0 (and newer) has less IDs than 1.57.0.0.0, a double check is needed.
		with open("#GOP_Files\\amd_gop_IDs.txt", 'r+') as id_file:
			for line in id_file:
				if line[:9] == ven_dev :
					id_in_gop = True
					#print("The ID %s is present in the GOP!\n" % ven_dev)
					break
		
		if id_in_gop :
			amd_file = "amd_gop.efirom"
		else :
			print(Style.BRIGHT + Fore.YELLOW + "  Warning! Your VBIOS ID %s doesn't exist in latest available GOP!\n" % ven_dev + 
			Fore.RESET + Style.NORMAL)
			ask = input("\nDo you still want to update GOP? Y for yes or any key for checking the ID in older 1.57.0.0.0 GOP: ")
			
			if ask.strip().upper() == "Y" :
				amd_file = "amd_gop.efirom"
				print("")
			else :
				with open("#GOP_Files\\amd_gop_IDs_1.57.0.0.0.txt", 'r+') as id_file:
					for line in id_file:
						if line[:9] == ven_dev :
							id_in_gop = True
							#print("The ID %s is present in the GOP!\n" % ven_dev)
							break
				print("")
				
				if id_in_gop :
					last_gop = last_amd_old ## This is only to have the proper updated version displayed.
					amd_file = "amd_gop_1.57.0.0.0.efirom"
				else :
					print(Style.BRIGHT + Fore.YELLOW + "  Warning! Your VBIOS ID %s doesn't exist in older GOP!\n" % ven_dev + 
					Fore.RESET + Style.NORMAL)
					ask = input("\nDo you still want to update GOP? A for %s, B for 1.57.0.0.0 or any key for exit: " % last_amd_new)
					ask = ask.strip().upper()
					
					if ask == "A" :
						amd_file = "amd_gop.efirom"
					elif ask == "B" :
						last_gop = last_amd_old ## This is only to have the proper updated version displayed.
						amd_file = "amd_gop_1.57.0.0.0.efirom"
					else :
						#f.close()
						sys.exit()
					
					print("")
		
		with open("#GOP_Files\\%s" % amd_file, 'rb') as amd_gop :
			gop_rom = amd_gop.read()
		
	elif gop_type == "Nvidia" and nv_type != "GXxxx" :
		
		#gop_type = "Nvidia"
		efi_id_off  = 0x20 ## might change it future versions
		efr_lst_off = 0x31 ## might change it future versions
		efi_lst_off = 0x4A ## might change it future versions
		
		## Make sure the Nvidia GOP is not customized. Only the last digit of variant ID should be non-zero.
		try :
			with open("#GOP_Files\\%s" % nv_file, 'rb') as nv_gop :
				gop_rom = nv_gop.read()
		except :
			print(Style.BRIGHT + Fore.RED + "  Unable to find a matching GOP! Please report it!\n" + 
			Fore.RESET + Style.NORMAL)
			sys.exit()
	
	elif pci_ven == "10DE" and (gop_type == "" or gop_type == "Nvidia") :
		
		gop_type = "Nvidia"
		efi_id_off  = 0x20 ## might change it future versions
		efr_lst_off = 0x31 ## might change it future versions
		efi_lst_off = 0x4A ## might change it future versions
		
		print(Style.BRIGHT + Fore.YELLOW + "  Warning! GOP type missing! Continue only if you know what you are doing!\n" + 
		Fore.RESET + Style.NORMAL)
		
		gpu_hint = nvidia_board()
		
		print(Style.BRIGHT + Fore.YELLOW + "  Product name = %s. This might (!!) be used to determine your GPU architecture.\n" % gpu_hint + 
		Fore.RESET + Style.NORMAL)
		
		print("\nDo you still want to update GOP? Select the number of your GPU architecture: \n\n")
		print("  1 = GT21x")
		print("  2 = GF10x")
		print("  3 = GF119")
		print("  4 = GK1xx")
		print("  5 = GM1xx")
		print("  6 = GM2xx")
		print("  7 = GP1xx")
		print("  8 = GK1xx_MXM")
		print("  9 = GM1xx_MXM")
		
		while True :
			ask = input("\n\nEnter choice: ")
			ask = ask.strip()
			
			if ask == "1" :
				last_gop = last_nv_GT21x
				nv_file = "nv_gop_GT21x.efirom"
				break
			elif ask == "2" :
				last_gop = last_nv_GF10x
				nv_file = "nv_gop_GF10x.efirom"
				break
			elif ask == "3" :
				last_gop = last_nv_GF119
				nv_file = "nv_gop_GF119.efirom"
				break
			elif ask == "4" :
				last_gop = last_nv_GK1xx
				nv_file = "nv_gop_GK1xx.efirom"
				break
			elif ask == "5" :
				last_gop = last_nv_GM1xx
				nv_file = "nv_gop_GM1xx.efirom"
				break
			elif ask == "6" :
				last_gop = last_nv_GM2xx
				nv_file = "nv_gop_GM2xx.efirom"
				break
			elif ask == "7" :
				last_gop = last_nv_GP1xx
				nv_file = "nv_gop_GP1xx.efirom"
				break
			elif ask == "8" :
				last_gop = last_nv_GK1xx_MXM
				nv_file  = "nv_gop_GK1xx_MXM.efirom"
				break
			elif ask == "9" :
				last_gop = last_nv_GM1xx_MXM
				nv_file  = "nv_gop_GM1xx_MXM.efirom"
				break
			else :
				print("\nWrong choice! Self destruct in 10, 9, 8, ...")
				#f.close()
				#sys.exit()
		
		print("")
		
		with open("#GOP_Files\\%s" % nv_file, 'rb') as nv_gop :
			gop_rom = nv_gop.read()
	
	elif gop_type[:4] == "Mac_" :
		efi_id_off  = 0x20 ## Might not always be true
		efr_lst_off = 0x31 ## Might not always be true
		efi_lst_off = 0x4A ## Might not always be true
		
		if not os.path.isfile("#GOP_Files\\%s" % mac_file) :
			print(Fore.RED + "File %s was not found!" % mac_file + Fore.RESET)
			sys.exit()
		
		with open("#GOP_Files\\%s" % mac_file, 'rb') as mac_gop :
			gop_rom = mac_gop.read()
	else :
		print(Style.BRIGHT + Fore.YELLOW + "Only AMD and Nvidia GOP supported!" + Fore.RESET + Style.NORMAL)
		#f.close()
		sys.exit()
	
	orom_end      = orom_start + orom_size
	orom_pci_last = orom_pcir_off + 0x15
	orom_last_img = ord(reading[orom_pcir_off + 0x15:orom_pcir_off + 0x16]) & 0x80
	end_img_old   = orom_end + efi_size ## This is only [IFR + ] ROM + EFI, not all sections.
	end_img_new   = orom_end + len(gop_rom)
	all_size      = len(reading)
	orom_cl_code  = reading[orom_pcir_off + 0xD:orom_pcir_off + 0x10]
	weird_npds_ps = False
	
	## Check for special images between ROM and EFI, but not in first container. Don't know why Nvidia is doing this.
	if efi_found and efi_begin != orom_end : # Already checked for weird data in rom_info, must be special images in between
		end_img_old   = efi_begin + efi_size
		end_img_new   = efi_begin + len(gop_rom)
		weird_npds_ps = True
	
	## TODO Special case for old type with EFI between ROM and special images with 55AA. Check after last image.
	## Check for other ROM images after ROM + EFI.
	if rom_info(reading[end_img_old:], "basic") :
		print(Style.BRIGHT + Fore.RED + "  There are other ROM images in this binary! Please report it!\n" + 
			Fore.RESET + Style.NORMAL)
	
	## Relocate the microcode.
	## Struct of microcode - https://github.com/torvalds/linux/blob/master/drivers/gpu/drm/radeon/atombios.h
	mc_reloc = False
	if pci_ven == '1002' and re.search(b'MCuC', reading) is not None :
		mcuc_list = re.finditer(b'MCuC', reading)
		
		for mcuc_match in mcuc_list :
			(mcuc_bgn, end_mcuc_match) = mcuc_match.span()
			
			mc_off = int.from_bytes(reading[mcuc_bgn - 8:mcuc_bgn - 4], 'little')
			
			if reading[mc_off:mc_off + 4] == b'MCuC' :
				
				## Nothing to do if microcode doesn't move.
				if end_img_new <= mc_off :
					print(Style.BRIGHT + Fore.YELLOW + "  AMD microcode will remain at the same offset.\n" + Fore.RESET + Style.NORMAL)
				else :
					mc_reloc  = True
					parm_size = int.from_bytes(reading[mc_off + 8:mc_off + 0x0A], 'little')
					code_size = int.from_bytes(reading[mc_off + 0x0A:mc_off + 0x0C], 'little')
					mc_size   = 0x10 + parm_size + code_size
					#print("%0.2X" % mc_size)
					mc_round  = mc_size + ((0x10 - mc_size % 0x10) & 0x0F)
					#print("%0.2X" % mc_round)
					mc_end    = mc_off + mc_round
					#new_mc_off = end_img_new + mc_off - end_img_old
					new_mc_off = end_img_new
					new_mc_end = new_mc_off + mc_round
					new_mc_bin = new_mc_off.to_bytes(4, 'little')
					print(Style.BRIGHT + Fore.YELLOW + "  AMD microcode will be relocated to offset 0x%0.2X.\n" % new_mc_off + Fore.RESET + Style.NORMAL)
					reading  = reading[:mcuc_bgn - 8] + new_mc_bin + reading[mcuc_bgn - 4:]#end_img_old] + reading[mc_off:mc_end]
				
				break
		
		else :
			print(Style.BRIGHT + Fore.YELLOW + "  AMD microcode pointer was found, but not its target!\n" + Fore.RESET + Style.NORMAL)
	
	## Fix first image for EFI pointing.
	if orom_last_img :
		
		print(Style.BRIGHT + Fore.YELLOW + "  Fixing last-image-bit in PCI Structure of Legacy ROM! \n" + Fore.RESET + Style.NORMAL)
		
		## Determine checksum byte
		if gop_type == "AMD" :
			imb_test = reading[orom_start + 0x1E:orom_start + 0x21]
			
			if imb_test == b'IBM' :
				ibm_end = orom_start + 0x21
				chk_is_last = False
			else :
				ibm_sig = re.search(br'\x49\x42\x4D', reading[:0xD0])
				
				if ibm_sig is not None :
					(start_ibm, ibm_end) = ibm_sig.span()
					chk_is_last = False
					print(Style.BRIGHT + Fore.YELLOW + "  Checksum byte of Legacy OROM at offset 0x%0.2X! \n" % ibm_end + 
					Fore.RESET + Style.NORMAL)
				else :
					ibm_end = orom_end - 1
					chk_is_last = True
			
			chk_int_old = ord(reading[ibm_end:ibm_end + 1])
			chk_off     = ibm_end
		
		elif gop_type == "Nvidia" :
			chk_int_old = ord(reading[orom_end - 1:orom_end])
			chk_off     = orom_end - 1
			chk_is_last = True
		
		## Determine if there is one container for all ROMs. EFI is most likely missing.
		if reading[orom_pcir_off + 0x20:orom_pcir_off + 0x24] == b'NPDE' :
			npde_size = int.from_bytes(reading[orom_pcir_off + 0x28:orom_pcir_off + 0x2A], 'little') * 0x200
			
			if orom_size != npde_size :
				print(Style.BRIGHT + Fore.RED + "  Different sizes in PCI structure and NPDE structure of Legacy ROM!\n" + 
				Fore.RESET + Style.NORMAL)
				
				orom_end_npde = orom_start + npde_size
				rom_test_npde = reading[orom_end_npde:orom_end_npde + 2]
				
				## If NPDE size is the right one, there is one container for all ROMs.
				if rom_test_npde in [b'\x55\xAA', b'\x56\x4E'] :
					print(Style.BRIGHT + Fore.YELLOW + "  The Legacy ROM appears to be a container for all images.\n" + 
					Fore.RESET + Style.NORMAL)
					print(Style.BRIGHT + Fore.YELLOW + "  Fixing last-image-bit in last special image.\n" + Fore.RESET + Style.NORMAL)
					nv_step      = orom_end_npde
					check_nv_ext = reading[nv_step:nv_step + 2]
					
					while check_nv_ext == b'\x56\x4E' or check_nv_ext == b'\x55\xAA' :
						nv_ext_size = int.from_bytes(reading[nv_step + 2:nv_step + 4], 'little') * 0x200
						#print(nv_ext_size)
						
						if nv_ext_size == 0 :
							npds_off    = int.from_bytes(reading[nv_step + 0x18:nv_step + 0x1A], 'little')
							npds_off    += nv_step
							nv_ext_size = int.from_bytes(reading[npds_off + 0x10:npds_off + 0x12], 'little') * 0x200
							#print(nv_ext_size)
						
						nv_step      += nv_ext_size
						check_nv_ext = reading[nv_step:nv_step + 2]
					
					lst_img_npds_off = npds_off + 0x15
					lst_npds_int_old = ord(reading[lst_img_npds_off:lst_img_npds_off + 1])
					lst_npds_int_new = int(lst_npds_int_old & 0x7F)
					lst_npds_bin_new = bytes([lst_npds_int_new])
					
					checksum_old = sum(bytearray(reading[orom_start:orom_end_npde - 1]))
					checksum_new = (checksum_old - lst_npds_int_old + lst_npds_int_new) & 0xFF
					chk_int_new  = 256 - checksum_new if checksum_new else 0
					chk_bin_new  = bytes([chk_int_new])
					
					reading = reading[:orom_end_npde - 1] + chk_bin_new + reading[orom_end_npde:lst_img_npds_off] + \
							lst_npds_bin_new + reading[lst_img_npds_off + 1:]
					
					npde_off_test    = npds_off + 0x20
					lst_img_npde_off = npds_off + 0x2A
					
					if reading[npde_off_test:npde_off_test + 4] == b'NPDE' :
						lst_npde_int_old = ord(reading[lst_img_npde_off:lst_img_npde_off + 1])
						lst_npde_int_new = int(lst_npde_int_old & 0x7F)		
						lst_npde_bin_new = bytes([lst_npde_int_new])
						
						reading = reading[:lst_img_npde_off] + lst_npde_bin_new + reading[lst_img_npde_off + 1:]
		
		checksum_old = sum(bytearray(reading[orom_start:orom_end])) ## sumbytes(reading[orom_start:orom_end])
		
		lst_int_old  = ord(reading[orom_pci_last:orom_pci_last + 1])
		lst_int_new  = int(lst_int_old & 0x7F)
		lst_bin_new  = bytes([lst_int_new])
		
		checksum_new = (checksum_old - chk_int_old - lst_int_old + lst_int_new) & 0xFF
		
		chk_int_new  = 256 - checksum_new if checksum_new else 0
		chk_bin_new  = bytes([chk_int_new])
		
		if chk_is_last :
			print(Style.BRIGHT + Fore.YELLOW + "  Using last byte for checksum! \n" + Fore.RESET + Style.NORMAL)
			new_gop = reading[:orom_pci_last] + lst_bin_new + reading[orom_pci_last + 1:orom_end - 1] + chk_bin_new
		else :
			print(Style.BRIGHT + Fore.YELLOW + "  Using AMD byte for checksum! \n" + Fore.RESET + Style.NORMAL)
			new_gop = reading[:chk_off] + chk_bin_new + reading[chk_off + 1:orom_pci_last] + lst_bin_new + reading[orom_pci_last + 1:orom_end]
	else :
		new_gop = reading[:orom_end]
	
	## Add special images between ROM and EFI. If they are present and not part of main container, nothing else to do.
	## The last-bit is already set in ROM and special images.
	if weird_npds_ps :
		new_gop += reading[orom_end:efi_begin]
	
	## Assembly a new image
	if gop_type == "Nvidia" or gop_type == "Mac_Nvidia" :
		## Nvidia has special images and special structures, needs more care.
		efi_id_old   = sum(bytearray(gop_rom[efi_id_off:efi_id_off + 4]))
		efi_lst_old  = ord(gop_rom[efi_lst_off:efi_lst_off + 1]) #sum(bytearray(gop_rom[efi_lst_off:efi_lst_off + 1]))
		efr_lst_old  = ord(gop_rom[efr_lst_off:efr_lst_off + 1]) #sum(bytearray(gop_rom[efr_lst_off:efr_lst_off + 1]))
		checksum_old = sum(bytearray(gop_rom[:-1]))
		
		## Check if EFI is last image in Nvidia VBIOS. Change the bit in NPDE.
		check_nv_ext = reading[end_img_old:end_img_old + 2]
		if check_nv_ext == b'\x56\x4E' or check_nv_ext == b'\x55\xAA' :
			print(Style.BRIGHT + Fore.YELLOW + "  EFI is NOT last image!\n" + Fore.RESET + Style.NORMAL)
			efi_lst_new = int(efi_lst_old & 0x7F)
			efr_lst_new = int(efr_lst_old & 0x7F)
			## Remove end padding from dumped images.
			nv_step  = end_img_old
			
			while check_nv_ext == b'\x56\x4E' or check_nv_ext == b'\x55\xAA' :
				nv_ext_size = int.from_bytes(reading[nv_step + 2:nv_step + 4], 'little') * 0x200
				#print(nv_ext_size)
				
				if nv_ext_size == 0 :
					npds_off = int.from_bytes(reading[nv_step + 0x18:nv_step + 0x1A], 'little')
					npds_off += nv_step
					nv_ext_size = int.from_bytes(reading[npds_off + 0x10:npds_off + 0x12], 'little') * 0x200
					#print(nv_ext_size)
				
				nv_step += nv_ext_size
				check_nv_ext = reading[nv_step:nv_step + 2]
			else :
				if check_nv_ext == b'' :
					## No extra data and no padding, so we can re-add special images as end data.
					end_data = reading[end_img_old:]
				else :
					print(Style.BRIGHT + Fore.YELLOW + "  Removing unnecessary end padding.\n" + Fore.RESET + Style.NORMAL)
					end_data = reading[end_img_old:nv_step] ## This is the normal situation, where only padding follows last special image.
					str_err  = "  Data after Nvidia special images! Please report it!\n"
					end_img_new += nv_step - end_img_old
					end_data = remove_padding(end_data, nv_step, end_img_new, all_size, str_err)
		
		else :
			print(Style.BRIGHT + Fore.YELLOW + "  EFI is last image.\n" + Fore.RESET + Style.NORMAL)
			efi_lst_new = int(efi_lst_old | 0x80)
			efr_lst_new = int(efr_lst_old | 0x80)
			## Remove end padding from dumped images.
			end_data = b''
			
			if end_img_old < all_size :
				print(Style.BRIGHT + Fore.YELLOW + "  Removing unnecessary end padding.\n" + Fore.RESET + Style.NORMAL)
				str_err  = "  Data after ROM and not part of Nvidia special images! Please report it!\n"
				end_data = remove_padding(end_data, end_img_old, end_img_new, all_size, str_err)
		
		efi_id_new      = sum(bytearray(orom_id_bin))
		efi_lst_bin_new = bytes([efi_lst_new])
		efr_lst_bin_new = bytes([efr_lst_new])
		
		print(Style.BRIGHT + Fore.YELLOW + "  Fixing ID, last-image-bit and checksum for EFI image.\n" + Fore.RESET + Style.NORMAL)
		
		if orom_old_type :
			print(Style.BRIGHT + Fore.YELLOW + "  Fixing last-image-bit in PCIR and NPDE for EFI image.\n" + Fore.RESET + Style.NORMAL)
		else :
			efr_lst_new     = efr_lst_old
			efr_lst_bin_new = bytes([efr_lst_new])
				
		checksum_new    = (checksum_old - efi_id_old - efr_lst_old - efi_lst_old + efi_id_new + efr_lst_new + efi_lst_new) & 0xFF
		efi_chk_int_new = 256 - checksum_new if checksum_new else 0
		efi_chk_bin_new = bytes([efi_chk_int_new])
				
		new_gop += gop_rom[:efi_id_off] + orom_id_bin + gop_rom[efi_id_off + 4:efi_id_off + 9] + efi_class_bin + \
					gop_rom[efi_id_off + 0xC:efr_lst_off] + efr_lst_bin_new + gop_rom[efr_lst_off + 1:efi_lst_off] + \
					efi_lst_bin_new + gop_rom[efi_lst_off + 1:-1] + efi_chk_bin_new + end_data
	
	else :
		## gop_type is "AMD"
		## AMD has no special images, from limited testing.
		print(Style.BRIGHT + Fore.YELLOW + "  Fixing ID for EFI image. No checksum correction is needed.\n" + Fore.RESET + Style.NORMAL)
		
		## Remove end padding from dumped images.
		end_data = b''
		
		## Standard error message for extra data
		str_err  = "  Data after ROM and not part of EFI! Please report it!\n"
		
		## If AMD microcode is present, check for extra data after it. Assume none before.
		if mc_reloc :
			str_err  = "  Data after microcode! Please report it!\n"
			end_img_old = mc_end
			end_img_new = new_mc_end
		
		if end_img_old < all_size :
			print(Style.BRIGHT + Fore.YELLOW + "  Removing unnecessary end padding.\n" + Fore.RESET + Style.NORMAL)
			#str_err  = "  Data after ROM and not part of EFI! Please report it!\n"
			end_data = remove_padding(end_data, end_img_old, end_img_new, all_size, str_err)
		
		## Add the microcode and any extra data.
		if mc_reloc :
			end_data = reading[mc_off:mc_end] + end_data
				
		new_gop += gop_rom[:efi_id_off] + orom_id_bin + gop_rom[efi_id_off + 4:efi_id_off + 9] + efi_class_bin + \
					gop_rom[efi_id_off + 0xC:] + end_data
	
	with open("%s_updGOP%s" % (fileName, fileExtension), 'wb') as my_gop :
		my_gop.write(new_gop)
	
	print(Style.BRIGHT + Fore.CYAN + "\nFile \"%s_updGOP%s\" with updated GOP %s was written!\n" % (fileName, fileExtension, last_gop) + 
	Fore.RESET + Style.NORMAL)

###########################################################